import datetime
from typing import Optional, Any

from psycopg.errors import RaiseException
from pydantic import BaseModel, Field

from util.util import get_cursor
from model.wapi.nd import BCD
from util.wapi_util import execute_wapi_function


# TODO: properly rename to WAPIAuth
# TODO: fix fields
class APIToken(BaseModel):
    gpk: str | Any = Field(default='')
    token: str | Any = Field(default='')
    login_name: str | Any = Field(default='')
    description: str | Any = Field(default="")
    id: Optional[int | Any] = Field(default=None)

    @staticmethod
    def from_token(cursor, token: str):
        if 'Bearer' in token:
            token = token.replace('Bearer ', '')
        try:
            mgr = Mgr.check_token(cursor, token)
            return APIToken(token=token, login_name=mgr.login_name)
        except RaiseException:
            # TODO: fixme
            return None


class Mgr(BaseModel):
    # TODO add more fields (and properly use Field(...))
    permission_list: Optional[list] = None
    bcds: Optional[list[BCD]] = Field(default=None)
    # db-fields
    login_name: str = Field()
    first_name: Optional[str] = Field(default='')
    last_name: Optional[str] = Field(default='')
    email: Optional[str] = Field(default=None)
    adm_lock: bool = Field(default=False)
    kit_status: Optional[int] = Field(default=None)
    allow_data_manipulation: bool = Field(default=True)
    description: Optional[str] = Field(default=None)
    kit_ou_short_name: Optional[str] = Field(default=None)
    is_svc: bool = Field(default=False)
    sub_id: Optional[str] = Field(default=None)
    svc_id: Optional[str] = Field(default=None)

    def create_session_token(self, conn, stay_logged_in=False):

        descr = 'Session-Token ({})'.format(datetime.datetime.now())

        cursor = get_cursor(conn)
        with conn.transaction():
            cursor.execute("SELECT out_token_text, out_token_gpk "
                           "FROM cntl.create_sess_auth(%(login_name)s, %(descr)s, %(stay_logged_in)s)",
                           {
                               'login_name': self.login_name,
                               'stay_logged_in': stay_logged_in,
                               'descr': descr
                           })
            res = cursor.fetchone()

        return APIToken(gpk=res['out_token_gpk'],
                        token=res['out_token_text'],
                        login_name=self.login_name,
                        description=descr)

    @staticmethod
    def check_token(conn, token: str | APIToken):
        cursor = get_cursor(conn)
        with conn.transaction():
            if isinstance(token, APIToken):
                token = token.token
            cursor.execute("select * from cntl.validate_jwt(in_token:=%(token)s)", {'token': token})

            result = cursor.fetchone()
            if result is None:
                return None

            return Mgr(**result)

    def get_bcds(self, conn) -> list[BCD]:
        if self.bcds is None:
            r = execute_wapi_function(conn, [
                {'idx': 'bcd_list', 'name': 'nd.bcd.list', 'old': {'is_own': True}},
            ], user=self.login_name)

            self.bcds = [BCD(**v) for v in r['bcd_list']]
        return self.bcds

    def has_bcd(self, conn, bcd: str | BCD) -> bool:
        if isinstance(bcd, BCD):
            bcd = bcd.name
        return bcd in [v.name for v in self.get_bcds(conn)]

    def get_permissions(self, conn):
        if self.permission_list is not None:
            return self.permission_list

        result = execute_wapi_function(conn, [
            {'name': 'cntl.mgr.list', 'old': {'is_self': True}},
            {'name': 'cntl.role.list', 'idx': 'direct_role_list', 'old': {
                'is_directly_assigned': True,
                'is_role': True,
            }},
            {'name': 'cntl.role.list', 'old': {'is_role': True}, 'idx': 'role_list', 'inner_join_ref': {
                'direct_role_list': 'api_func_cntl_role_hierarchy_gfk_list_is_superset'
            }},
            {'name': 'cntl.role.list', 'old': {'is_role': False}, 'idx': 'permission_list', 'inner_join_ref': {
                'role_list': 'api_func_cntl_role_hierarchy_gfk_list_is_superset'
            }}
        ], user=self.login_name)

        if 'permission_list' not in result.keys() or len(result['permission_list']) == 0:
            self.permission_list = []
            return self.permission_list

        self.permission_list = list(map(lambda x: x['fq_name'], result['permission_list']))
        return self.permission_list

    def has_permission(self, conn, permission: str):
        return permission in self.get_permissions(conn)
