import json
import logging

import psycopg

from model.settings import settings
from util.util import get_cursor

logger = logging.getLogger('uvicorn.error')


def transaction_result_to_stmt_dict(transaction_result: dict) -> dict:
    r = {}
    for res in transaction_result:
        idx = res['out_stmt_idx']
        if idx is not None:
            if r.get(idx, None) is None:
                r[idx] = []
            if res['out_obj_dict'] is not None:
                r[idx].append(res['out_obj_dict'])
    return r


def execute_wapi_function(conn, request: list[dict[str, str | dict]], user: str = None,
                          dry_mode: bool = True,
                          superuser: bool = False) -> dict:
    # for proper handling as a result-dict it is necessary for every statement to have an idx
    #   it is assumed for statements where the same resource is queried multiple times, only at most one of those
    #   statements to have no idx (as the others are required for joins, etc. anyways)
    for stmt in request:
        if 'idx' not in stmt.keys():
            stmt['idx'] = stmt['name'].replace('.', '_')

    # if any statement is not a list statement, we need to use a higher isolation level
    # this is just a heuristic, but it should work in most cases and it only creates false positives.
    maybe_contains_dml = any(not stmt['name'].endswith('.list') for stmt in request)

    if user is None and superuser:
        user = settings.wapi_netvs_superuser

    old_isolation_level = conn.isolation_level
    conn.isolation_level = psycopg.IsolationLevel.READ_COMMITTED
    if maybe_contains_dml:
        conn.isolation_level = psycopg.IsolationLevel.SERIALIZABLE

    query = """
               select * from wapi_4_1.ta_handler(
                 in_login_name => %(login_name)s,
                 in_http_rq_mthd => 'POST',
                 in_http_rq_body => %(rq)s,
                 in_report_stmt_pos => true,
                 in_language_tag=>'de',
                 in_is_dry_mode => %(dry_mode)s
               );
            """
    result = None
    try:
        cursor = get_cursor(conn)
        with conn.transaction():
            cursor.execute(query, {
                'login_name': user,
                'rq': json.dumps(request),
                'dry_mode': dry_mode,
            })
            result = cursor.fetchall()
    except Exception as e:
        logger.error('Error in executing statement:\n\n{stmt}\n\n'.format(stmt=json.dumps(request, indent=2)))
        raise e
    finally:
        conn.isolation_level = old_isolation_level

    if result is None:
        return {}

    return transaction_result_to_stmt_dict(result)
